#!/usr/bin/perl
###############################################################################
#    Copyright (C) 2002-2204 by Eric Gerbier
#    Bug reports to: eric.gerbier@tutanota.com
#    $Id$
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
###############################################################################
# afick_format is designed to read afick text report and convert it into
# others format (html, xml ...)
#
# this code is still in alpha stage
###############################################################################

use strict;
use warnings;
use Pod::Usage;
use English '-no_match_vars';
use Getopt::Long;    # option analysis

# debuggging
# use diagnostics;
# use Data::Dumper;
# use Carp qw(cluck);	# debugging

use File::Basename;    # for path
my $dirname = dirname($PROGRAM_NAME);
require $dirname . '/afick-common.pl';

my $EMPTY = Afick::Constant->EMPTY;

###############################################################################
#                     global variables
###############################################################################

my $Version = '0.6';

#############################################################
sub version($) {
	my $version = shift @_;
	print "\n";
	print
"afick_format : another file integrity checker logfile converter\nversion $version\n";
	return;
}

#############################################################
sub header_xml () {
	return <<'EOF_HXML';
<?xml version="1.0" standalone="yes" ?>
<!-- generated by afick -->
<afick>
EOF_HXML
}
#######################################################
sub header_html () {
	return <<'EOF_HHTML';
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
<html>
  <head>
        <meta name="generator"  content="afick">
         <title>afick result</title>
 </head>
 <body>
EOF_HHTML
}
#######################################################
sub footer_html () {
	return << 'EOF_FHTML';
  </body>
</html>
EOF_FHTML
}
#######################################################
sub footer_xml () {
	return <<'EOF_FXML';
</afick>
EOF_FXML
}
#############################################################
sub footer_txt () {
	return $EMPTY;
}
#############################################################
sub header_txt () {
	return $EMPTY;
}
#############################################################
sub check_format($) {
	my $f_in = shift @_;

	my %formats = (
		'txt'  => 1,
		'html' => 2,
		'xml'  => 3,
	);
	return exists $formats{$f_in};
}
#############################################################
# remove html tags
sub remove_tag($) {
	my $rlog_in = shift @_;

	my @log_out;

	foreach my $ligne ( @{$rlog_in} ) {
		chomp $ligne;
		$ligne =~ s/<[^>]*>//g;
		next if ( $ligne =~ m/^\s*$/ );
		push @log_out, $ligne;
	}

	return @log_out;
}
#############################################################
sub txt2xml($) {
	my $rlog_in = shift @_;

	my $log = Afick::Log->new();
	$log->parse_array($rlog_in);

	print header_xml();
	print "<metadata>\n";

	# run data
	my %run = $log->get_run();
	print "<run>\n";
	foreach my $key ( sort keys %run ) {
		print "<$key>$run{$key}</$key>\n";
	}
	print "</run>\n";

	# warnings and critical
	my @warnings = $log->get_warnings();
	print "<warnings>\n";
	foreach my $w (@warnings) {
		print "<warning>$w</warning>\n";
	}
	print "</warnings>\n";

	# config
	my %config = $log->get_config();
	print "<config>\n";
	foreach my $key ( sort keys %config ) {
		print "<$key>$config{$key}</$key>\n";
	}
	print "</config>\n";
	print "</metadata>\n";

	# data
	print "<data>\n";
	foreach my $type ( sort $log->get_data_types_k() ) {
		print "<$type" . "s>\n";
		foreach my $name ( sort $log->get_data_names_k($type) ) {
			print "<$type>\n";
			print "<name>$name</name>\n";
			foreach my $field ( sort $log->get_data_fields_k( $type, $name ) ) {
				print "<$field>\n";
				my $value = $log->get_data_field( $type, $name, $field );
				if ( $log->is_dual_field($value) ) {
					my ( $old, $new ) = $log->split_dual_field($value);
					print "<old>$old</old><new>$new</new>\n";
				}
				else {
					print $value;
				}
				print "</$field>\n";
			}
			print "</$type>\n";
		}
		print "</$type" . "s>\n";
	}
	print "</data>\n";

	print footer_xml();
	return;
}
#############################################################
sub txt2html($) {
	my $rlog_in = shift @_;

	my $log = Afick::Log->new();
	$log->parse_array($rlog_in);

	print header_html();
	print "<hr>\n";

	# summary
	print "<a href='#warnings'>warnings</a><br />\n";
	print "<a href='#run'>run</a><br />\n";
	print "<a href='#config'>config</a><br />\n";
	print "<a href='#data'>data</a><br />\n";
	foreach my $type ( sort $log->get_data_types_k() ) {
		print "<a href='#$type'>$type</a><br />\n";
	}
	print '<hr />';

	# warnings and criticals
	my @warnings = $log->get_warnings();
	print "<h2><a name='warnings'>warnings</a></h2>\n<font color=red><ul>";
	foreach my $w (@warnings) {
		print "<li>$w</li>\n";
	}
	print "</ul></font>\n";

	# run data
	my %run = $log->get_run();
	print "<h2><a name='run'>run</a></h2>\n<font color=blue><ul>";
	foreach my $key ( sort keys %run ) {
		print "<li>$key : $run{$key}</li>\n";
	}
	print "</ul></font>\n";

	# config
	my %config = $log->get_config();
	print "<h2><a name='config'>config</a></h2>\n<font color=blue><ul>";
	foreach my $key ( sort keys %config ) {
		print "<li>$key : $config{$key}</li>\n";
	}
	print "</ul></font>\n";

	# data
	print "<h2><a name='data'>data</a></h2>\n";
	foreach my $type ( sort $log->get_data_types_k() ) {
		print "<h3><a name='$type'>$type</a></h3>\n";
		foreach my $name ( sort $log->get_data_names_k($type) ) {
			print "<h4>$name</h4>\n<ul>";
			foreach my $field ( sort $log->get_data_fields_k( $type, $name ) ) {
				print "<li>$field : ";
				my $value = $log->get_data_field( $type, $name, $field );
				if ( $log->is_dual_field($value) ) {
					my ( $old, $new ) = $log->split_dual_field($value);
					print "old=$old new=$new";
				}
				else {
					print $value;
				}
				print "</li>\n";
			}
			print "</ul>\n";
		}
	}

	print '</pre>';
	print "<hr>\n";
	print footer_html();
	return;
}
#############################################################
# try to guess file format
sub get_format($) {
	my $rlog_in = shift @_;

	my $format;

	foreach my $first_line ( @{$rlog_in} ) {

		## no critic (ProhibitCascadingIfElse)
		# skip WARNING lines
		if ( $first_line =~ m/WARNING/ ) {
			next;
		}
		elsif ( $first_line =~ m/^#/ ) {
			$format = 'txt';
			last;
		}
		## no critic(ProhibitEscapedMetacharacters)
		elsif ( $first_line =~ m/<\?xml/i ) {
			$format = 'xml';
			last;
		}
		elsif ( $first_line =~ m/html/i ) {
			$format = 'html';
			last;
		}
		else {
			last;
		}
	}
	return $format;
}
#############################################################
#                          main
#############################################################

$OUTPUT_AUTOFLUSH = 1;

# variables for parameter analysis
my %opt;
Getopt::Long::Configure('no_ignore_case');
if (
	!GetOptions(
		\%opt,

		# afickonfig options
		'help|h',
		'man',
		'version|V',
		'verbose|v',
		'output_format|o=s',
	)
  )
{
	pod2usage('incorrect option');
}

if ( exists $opt{'help'} ) {

	# -h : help
	pod2usage(1);
}
elsif ( exists $opt{'man'} ) {
	pod2usage( -verbose => 2 );
}
elsif ( exists $opt{'version'} ) {

	# -V : version
	version($Version);
	exit;
}

if ( exists $opt{'verbose'} ) {
	Afick::Msg->set_verbose( $opt{'verbose'} );
}

my $output_format = $opt{'output_format'};
if ($output_format) {

	if ( !check_format($output_format) ) {
		die "bad format $output_format\n";
	}
}
else {
	pod2usage('output format is mandatory');
}

# read all input file from stdin
my @log_in       = <>;
my $input_format = get_format( \@log_in );

## no critic (ProhibitCascadingIfElse)
if ( !defined $input_format ) {
	die "unknown input format\n";
}
elsif ( $input_format eq $output_format ) {

	# nothing to do
	print $_ . "\n" foreach (@log_in);
	exit;
}
elsif ( $input_format eq 'txt' ) {

	# nothing to do
}
elsif ( $input_format eq 'html' ) {
	@log_in = remove_tag( \@log_in );
}
elsif ( $input_format eq 'xml' ) {
	Afick::Msg->warning('not yet written');
	exit;
}
else {
	Afick::Msg->warning("strange bad input format : $input_format");
	exit;
}
## use critic

# convert to dest format
if ( $output_format eq 'html' ) {
	txt2html( \@log_in );
}
elsif ( $output_format eq 'xml' ) {
	txt2xml( \@log_in );
}
else {
	print $_ . "\n" foreach (@log_in);
}

__END__

=head1 NAME

afick_format - a tool to convert Afick's logfile format

=head1 DESCRIPTION

C<afick_format> is a tool to convert afick's log between different format :
txt (native format), html, xml. It can be used to have a better display than
the original one.

=head1 SYNOPSIS

afick_format.pl  [L<options|/OPTIONS>] logfile

=head1 REQUIRED ARGUMENTS

the output format

=head1 OPTIONS

options are used to control afickconfig

=over 4

=item --help|-h

Output help information and exit.

=item --man

Output full help information and exit.

=item --version|-V

Output version information and exit.

=item --verbose|-v

add debugging messages

=item --output_format|-o

can be txt, html or xml

=back

=head1 USAGE

C<afick_format.pl -o html /var/log/afick/afick.log>

=head1 NOTES

this program only use perl and its standard modules.

=head1 SEE ALSO

=for html
<a href="afick.conf.5.html">afick.conf(5)</a> for the configuration file syntax
<br>
<a href="afick-tk.1.html">afick-tk(1)</a> for the graphical interface
<br>
<a href="afick.1.html">afick(1)</a> for the command-line interface
<br>
<a href="afickonfig.1.html">afickonfig(1)</a> for a tool to change afick's configuration file
<br>
<a href="afick_archive.1.html">afick_archive(1)</a> for a tool to manage archive's reports
<br>
<a href="afick_learn.1.html">afick_learn(1)</a> for a learning tool

=for man
\fIafick.conf\fR\|(5) for the configuration file syntaxe
.PP
\fIafick\-tk\fR\|(1) for the graphical interface
.PP
\fIafick\fR\|(1) for the command-line interface
.PP
\fIafickonfig\fR\|(1) for a tool to change afick's configuration file
.PP
\fIafick_archive\fR\|(1) for a tool to manage archive's reports
.PP
\fIafick_learn\fR\|(1) for a learning tool

=head1 DIAGNOSTICS

the input format is analysed and we display some warnings if it does not match
what we are waiting for

=head1 EXIT STATUS

0 if all is ok

=head1 CONFIGURATION

it does not use afick's configuration file

=head1 DEPENDENCIES

perl, afick common library

=head1 INCOMPATIBILITIES

none

=head1 BUGS AND LIMITATIONS

input format can be txt or html, not xml
output format can be txt, html, or xml

=head1 LICENSE AND COPYRIGHT

Copyright (c) 2002 Eric Gerbier
All rights reserved.

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option)
any later version.

=head1 AUTHOR

Eric Gerbier

you can report any bug or suggest to eric.gerbier@tutanota.com
