#!/usr/bin/env perl -T
###############################################################################
#
#    Copyright (C) 2002-2023 by Eric Gerbier
#    Bug reports to: eric.gerbier@tutanota.com
#    $Id$
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
###############################################################################
# this classe is designed to implement package database access
###############################################################################
# I use special naming for references :
# $r_name : for a ref to scalar
# $rh_name : for a ref to hashes
# $ra_name : for a ref to array

package Afick::Deb;

use strict;
use warnings;

use English qw(-no_match_vars);

use base qw(Exporter);
## no critic (ProhibitAutomaticExportation)
our ( $VERSION, @EXPORT, );
$VERSION = '1.1';
@EXPORT  = qw( );

# optional modules
# ###################

# other afick modules
# #####################
use Afick::Constant;
use Afick::Msg;
use Afick::Gen;    # search_path

my $Dbcmd   = 'dpkg';
my $F_dbcmd = search_path($Dbcmd);    # full path to rpm command

## no critic (ProhibitBacktickOperators)
###############################################################
# constructor
sub new($) {
	my $classe = shift @_;

	my $self = {};
	bless $self, $classe;
	$self->_init();
	return $self;
}

###############################################################
# reset internal variables
sub _init($) {
	my $self = shift @_;

	# internal cache
	#$self->{'cache'} = undef;

	return;
}
################################################################
# is used to check if this database is installed and in use
sub check($) {
	my $self = shift @_;

	# check if dpkg is installed
	if ($F_dbcmd) {
		my @out = `$F_dbcmd -l $Dbcmd 2>&1 `;

		my @lines = grep { /^ii/ } @out;
		return $lines[0];
	}
	else {
		# command not found
		return;
	}
}
################################################################
# convert a filename to a package name
sub file2pac($$) {
	my $self     = shift @_;
	my $filename = shift @_;

	return unless ($F_dbcmd);
	return unless ($filename);
	return unless ( -e $filename );

	Afick::Msg->debug("(file2pac) $filename");

	# dpkg: /usr/bin/dpkg
	# python3-pycurl, python3-notify2: /usr/share/doc
	my $out = `$F_dbcmd -S $filename 2>/dev/null`;
	if ( $out =~ m/(.*): .*/ ) {
		my $pacs = $1;
		my @pacs = split m/[,]/, $pacs;
		my $pac  = $pacs[0];
		return $pac;
	}

	return;
}
################################################################
# return the list of files in the given package
sub list_pac_files($$) {
	my $self = shift @_;
	my $pac  = shift @_;

	return unless ($F_dbcmd);
	return unless $pac;
	Afick::Msg->debug("(list_pac_files) $pac");

	my @list = `$F_dbcmd -L $pac`;
	chomp @list;
	return @list;
}
################################################################
# return a hash of modified files against package database
sub list_changed_files($$) {
	my $self = shift @_;
	my $pac  = shift @_;

	return unless ($F_dbcmd);
	Afick::Msg->debug("(list_changed_files) $pac");

	my %h;
	my $cmd = "$F_dbcmd -V $pac 2>/dev/null";

	# ??5?????? c /etc/afick.conf
	## no critic (ProhibitTwoArgOpen)
	if ( open my $fh, "$cmd |" ) {
		while (<$fh>) {
			my @tab      = split q{ };
			my $changes  = $tab[0];
			my $filename = $tab[-1];
			chomp $filename;
			$h{$filename} = $changes;
		}
		## no critic (RequireCheckedClose,RequireCheckedSyscalls)
		close $fh;
	}
	else {
		Afick::Msg->warning("(verify_pac) can not open $cmd : $ERRNO");
	}
	return %h;
}
################################################################

1;
__END__

=head1 NAME

Afick::Deb

=head1 DESCRIPTION

=head1 DIAGNOSTICS

=head1 USAGE

use Afick::Deb;

my $pm = Afick::Deb->new();

$res = $pm->file2pac('/etc/fstab');

my %h = $pm->verify_pac(bash);

=head1 OPTIONS

=head1 REQUIRED ARGUMENTS

=head1 EXIT STATUS

=head1 CONFIGURATION

=head1 DEPENDENCIES

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 LICENSE AND COPYRIGHT

Copyright (c) 2002 Eric Gerbier
All rights reserved.

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option)
any later version.

=head1 AUTHOR

Eric Gerbier

you can report any bug or suggest to eric.gerbier@tutanota.com

