#!/usr/bin/env perl -T
###############################################################################
#
#    Copyright (C) 2002-2004 by Eric Gerbier
#    Bug reports to: eric.gerbier@tutanota.com
#    $Id$
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
###############################################################################
# this classes implements data and methods about a file
# as seen for afick
###############################################################################
# I use special naming for references :
# $r_name : for a ref to scalar
# $rh_name : for a ref to hashes
# $ra_name : for a ref to array

package Afick::Object;

use strict;
use warnings;

use English qw(-no_match_vars);

use Afick::Gen;    # file_type_h get_time
use Afick::Tst;    # is_acl
use Afick::Msg;

#use Afick::WinAcl;

our ($VERSION);
$VERSION = '0.3';

# database internal separator
# for join
## no critic(ProhibitConstantPragma)
use constant Sep => q{|};
## use critic

# for split
my $Sepmeta = quotemeta Sep;

# value give the order for display
my %Fields = (
	'filetype' => 0,     # computed file type
	'name'     => 1,     # file name
	'md5'      => 2,     # md5 checksum
	'device'   => 3,     # device ident
	'inode'    => 4,     # inode number
	'filemode' => 5,     # file mode  (type and permissions)
	'links'    => 6,     # number of links
	'uid'      => 7,     # numeric user ID of file's owner
	'gid'      => 8,     # numeric group ID of file's owner
	'filesize' => 9,     # file size in bytes
	'blocs'    => 10,    # actual number of system-specific blocks allocated
	'atime'    => 11,    # last access time in seconds since the epoch
	'mtime'    => 12,    # last modify time in seconds since the epoch
	'ctime'    => 13,    # inode change time in seconds since the epoch
	'acl'      => 14,    # acl (not used)
	'sha1'     => 15,
	'sha256'   => 16,
	'sha512'   => 17,
);

my $Utc_time = 0;        # store flag on class data

###############################################################
# constuctor
sub new($) {
	my $classe = shift @_;

	my $self = {};
	bless $self, $classe;
	$self->_init();
	return $self;
}

###############################################################
# initialize all fields
sub _init($) {
	my $self = shift @_;

	foreach my $field ( $self->fields() ) {
		$self->$field(0);
	}
	return;
}
###############################################################
# to be used as class method
sub set_utc_time($$) {
	my $self     = shift @_;
	my $utc_time = shift @_;    # a binary value

	$Utc_time = $utc_time;

	return;
}
###############################################################
sub get_utc_time($) {
	my $self = shift @_;

	return $Utc_time;
}
###############################################################
# build accessors (cf design patterns)
## no critic (ProhibitNoStrict)
no strict 'refs';
foreach my $field ( keys %Fields ) {
	*{$field} = sub : lvalue {
		my $self = shift @_;

		if (@_) {
			$self->{$field} = shift @_;    # set
			return;
		}
		else {
			return $self->{$field};        # get
		}
	};
}
use strict 'refs';
###############################################################
# return list of internal fields names
# no order
sub fields($) {
	my $self = shift @_;

	return keys %Fields;
}
###############################################################
# return the the list of internal fields sorted
sub sort_fields($) {
	my $self = shift @_;

	my @list = sort { $Fields{$a} <=> $Fields{$b} } keys %Fields;
	return @list;
}
###############################################################
# old database separator
sub sep($) {
	return Sep;
}
###############################################################
# get file_type from database
sub file_type ($) {
	my $self = shift @_;

	my $mode = $self->filemode();
	return file_type_h($mode);
}
###############################################################
#sub is_lnk($) {
#	my $self = shift @_;
#
#	return Afick::Gen::S_ISLNK( $self->filemode() );
#}
###############################################################
# old internal data flat storage format :
#   0          1       2      3       4      5    6
# $md5_sum, $device, $inode, $mode, $link, $uid, $gid,
#  7       8      9       10      11       12
# $size, $bloc, $atime, $mtime, $ctime, $sha1_sum, 0;
#
# fill data from database old flat storage format
# (read old database record : Afick::Backend->get)
sub from_oldflat($$$) {
	my $self = shift @_;
	my $key  = shift @_;    # file name
	my $dat  = shift @_;    # old flat storage format

	# clear data
	$self->_init();

	my $sep = $self->sep();
	Afick::Msg->warning("empty data for $key") if ( !$dat );
	Afick::Msg->warning("bad format for $key : $dat")
	  unless ( $dat =~ m/$sep/ );

	my @data = split /$Sepmeta/, $dat;

	$self->md5( $data[0] );
	$self->device( $data[1] );
	$self->inode( $data[2] );
	$self->filemode( $data[3] );
	$self->links( $data[4] );
	$self->uid( $data[5] );
	$self->gid( lc $data[6] );    # see WinAcl, force sid in lower case
	$self->filesize( $data[7] );
	$self->blocs( $data[8] );
	$self->atime( $data[9] );
	$self->mtime( $data[10] );
	$self->ctime( $data[11] );
	$self->sha1( $data[12] );

	$self->filetype( $self->file_type() );
	$self->name($key);
	return;
}

# --------------------------------------------------
# export data to old flat format
sub to_oldflat($) {
	my $self = shift @_;

	my @tab = (
		$self->md5(),   $self->device(), $self->inode(), $self->filemode(),
		$self->links(), $self->uid(),    $self->gid(),   $self->filesize(),
		$self->blocs(), $self->atime(),  $self->mtime(), $self->ctime(),
		$self->sha1(),  0,
	);
	return join Sep, @tab;
}

# --------------------------------------------------
sub is_oldflat($$) {
	my $self = shift @_;
	my $dat  = shift @_;    # old flat storage format

	my @data = split /$Sepmeta/, $dat;
	return ( scalar @data > 10 );
}

###############################################################
# initialise from a hash (file_info sub)
sub from_hash($$$) {
	my $self = shift @_;
	my $r_h  = shift @_;

	# clear data
	$self->_init();

	foreach my $key ( keys %{$r_h} ) {
		if ( exists $Fields{$key} ) {

			$self->$key( $r_h->{$key} );
		}
		else {
			Afick::Msg->warning("unknown key $key");
		}
	}

	$self->filetype( $self->file_type() );
	return;
}

# ----------------------------------------------------------
# return a hash
sub to_hash($) {
	my $self = shift @_;

	my %h;
	foreach my $field ( $self->fields() ) {
		$h{$field} = $self->{$field};
	}
	return %h;
}

# return a sorted array
# of data in human format
sub to_csv($) {
	my $self = shift @_;

	my @tab;
	foreach my $field ( $self->sort_fields() ) {
		push @tab, $self->display($field);
	}
	return @tab;
}

# to be used by export_raw function in Plugins.pm
sub export_raw($) {
	my $self = shift @_;

	my @tab;
	foreach my $field ( $self->sort_fields() ) {
		push @tab, $self->{$field};
	}
	return @tab;
}

###############################################################
# print all fields as raw data
sub debug($) {
	my $self = shift @_;

	foreach my $field ( $self->fields() ) {
		print "$field = " . $self->$field() . "\n";
	}
	return;
}
###############################################################
# convert a field from database in a human way
sub display($$) {
	my $self = shift @_;
	my $elem = shift @_;    # field id

	## no critic (ProhibitCascadingIfElse)
	if ( $elem eq 'device' ) {

		# device is in hexa, ex 309 == major 3, minor 9 == /dev/hda9
		return sprintf '%lx', $self->device();
	}
	elsif ( $elem eq 'filemode' ) {

		# perm is in octal
		return print_oct( $self->filemode() );
	}

	elsif ( ( $elem eq 'gid' ) and is_acl() ) {

		# gid on windows are acl
		## no critic (ProhibitNoisyQuotes)
		return join ';', Afick::WinAcl::split_acl( $self->gid() );
	}
	elsif ( $elem eq 'md5' ) {
		return print_digest( $self->md5() );
	}
	elsif ( $elem eq 'sha1' ) {
		return print_digest( $self->sha1() );
	}
	elsif ( $elem eq 'sha256' ) {
		return print_digest( $self->sha256() );
	}
	elsif ( $elem eq 'sha512' ) {
		return print_digest( $self->sha512() );
	}
	elsif (( $elem eq 'atime' )
		or ( $elem eq 'mtime' )
		or ( $elem eq 'ctime' ) )
	{

		# dates
		if ( $self->$elem() ) {
			my $date = get_time( $self->$elem(), $self->get_utc_time() );
			return $date;
		}
		else {
			return 0;
		}
	}
	else {
		return $self->$elem();
	}
	## use critic
}
#######################################################
# display_all
# convert in a human way all the record
# looks like internal storage
sub display_all($) {
	my $self = shift @_;

	my $output = join Sep, $self->to_csv();

	return $output;
}
#######################################################
# default array of file info
sub def_info() {
	return ( 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 );
}
#######################################################
# a new constructor
# file_info :
# build a complete info on the given file
# except checksum (too expensive)
## no critic (ProhibitExcessComplexity)
sub file_info($$$) {
	my $classe   = shift @_;
	my $name     = shift @_;    # file name
	my $masq     = shift @_;    #  attributes
	my $date_ref = shift @_;
	my $config   = shift @_;    # Afick::Cfg
	my $report   = shift @_;    # Afick::Report
	my $dbm      = shift @_;    # database type

	my $self = {};
	bless $self, $classe;

	# build a hash to initiate Afick::Object
	my %h = ( name => $name );

	my @fileinfo = lstat $name;
	if ( !@fileinfo ) {
		if ( -e $name ) {
			Afick::Msg->warning("(fileinfo) can not get lstat on $name");
		}
		else {
			Afick::Msg->debug("(fileinfo) temporary file $name");
		}

		# we must return an entry ...
		# so we fill it with null values
		@fileinfo = def_info();
	}

	# for afick-tk progress text
	Afick::Msg->progress($name);
	Afick::Msg->debug( $name . Afick::Object->sep(), 3 );

	my $is_link = Afick::Gen::S_ISLNK( $fileinfo[2] );
	my $fsize   = $fileinfo[7];
	$h{'filesize'} = $fsize;

	#my $md5_sum = 0;
	#my $sha_sum = 0;
	if ( -f $name ) {

		if ($fsize) {

			# just compute checksum on not empty files

			# test on file size to reduce cpu cost
			# if degraded mode, pass file size
			my $degraded          = 0;
			my $max_checksum_size = $config->get_directive('max_checksum_size');
			if ( ($max_checksum_size) and ( $fsize > $max_checksum_size ) ) {
				$degraded = $max_checksum_size;
				$report->inc_nb_degraded();
			}

			# tell checksum to read link's name, not link's file
			my $flag_symlink =
			  $is_link && ( not $config->get_directive('follow_symlinks') );

			if ( $masq =~ m/2/ ) {
				$h{'sha512'} = sha512sum( $name, $degraded, $flag_symlink );
			}
			elsif ( $masq =~ m/6/ ) {
				$h{'sha256'} = sha256sum( $name, $degraded, $flag_symlink );
			}
			elsif ( $masq =~ m/1/ ) {
				$h{'sha1'} = sha1sum( $name, $degraded, $flag_symlink );
			}
			if ( $masq =~ m/5/ ) {
				$h{'md5'} = md5sum( $name, $degraded, $flag_symlink );
			}
		}
		else {

			# empty file : skip it
		}
	}

	# See the perl "stat" command or man stat for more information.
	$h{'device'} = $fileinfo[0];
	$h{'inode'}  = $fileinfo[1];

	# file mode are different on unix and windows
	# user/group/other and acl
	# but we use mode to get file type, so keep this field as it
	$h{'filemode'} = $fileinfo[2];
	$h{'links'}    = $fileinfo[3];
	$h{'uid'}      = $fileinfo[4];

	my $gid;
	if ( is_microsoft() ) {

		# gid is unused on windows
		# it is replaced by the acess control list
		$gid = Afick::WinAcl::winacl( $name, $dbm );
	}
	else {

		# unix
		$gid = $fileinfo[5];
	}
	$h{'gid'} = $gid;

	$h{'atime'} = $fileinfo[8];
	my $mtime = $h{'mtime'} = $fileinfo[9];
	my $ctime = $h{'ctime'} = $fileinfo[10];
	$h{'blocs'} = $fileinfo[12];

	$self->from_hash( \%h );
	Afick::Msg->debug( $self->display_all(), 3 );

# Each file gets a pipe-delimited entry, with format as follows:
#
# field 0 = md5 checksum hash
# field 1 = device number of filesystem
# field 2 = inode number
# field 3 = file mode (type/permissions)
# field 4 = number of hard links to file
# field 5 = uid of the file
# field 6 = gid of the file
# field 7 = total size of file, in bytes
# field 8 = actual number of blocks allocated
# field 9 = last access time in seconds since the epoch
# field 10 = last modify time (since epoch)
# field 11 = inode change time
# field 12 = sha checksum hash
# we use a flat format here because the dbm database force us to store only this
# see also split_record sub to unpack the data
#	my $entry = join Sep,
#	  $md5_sum, $device, $inode, $mode,  $link,  $uid,     $gid,
#	  $fsize,   $bloc,   $atime, $mtime, $ctime, $sha_sum, 0;

	#	Afick::Msg->debug( display_all($entry), 3 );

	# !! seems to be a problem with -M/-C on dangling links
	if ( $config->get_directive('running_files') and ( !$is_link ) ) {
		my $cur_time = time;
		if ( $masq =~ m/m/ ) {
			if ( $mtime < $date_ref ) {

				# ok (before afick scan)
			}
			elsif ( $mtime > $cur_time ) {
				Afick::Msg->warning("(file_info) $name modified in the future");
			}
			else {
				Afick::Msg->warning(
"(file_info) $name has been modified during the afick scan( after $date_ref)"
				);
			}
		}
		if ( $masq =~ m/c/ ) {
			if ( $ctime < $date_ref ) {

				# ok (before afick scan)
			}
			elsif ( $ctime > $cur_time ) {
				Afick::Msg->warning(
					"(file_info) inode of $name modified in the future");
			}
			else {
				Afick::Msg->warning(
"(file_info) inode of $name has been modified during the afick scan (after $date_ref)"
				);
			}
		}
	}

	# return $entry;
	return $self;
}
#######################################################
# transform scalar masq in a masq array
sub masq2hash($) {
	my $masq = shift @_;    # internal masq

	my %masq_array = ();

	$masq_array{'md5'}      = 1 if ( $masq =~ m/5/ );    # md5 checksum
	$masq_array{'device'}   = 1 if ( $masq =~ m/d/ );    # device
	$masq_array{'inode'}    = 1 if ( $masq =~ m/i/ );    # inode
	$masq_array{'filemode'} = 1 if ( $masq =~ m/p/ );    # perms
	$masq_array{'links'}    = 1 if ( $masq =~ m/n/ );    # links
	$masq_array{'uid'}      = 1 if ( $masq =~ m/u/ );    # uid
	$masq_array{'gid'}      = 1 if ( $masq =~ m/g/ );    # gid
	$masq_array{'filesize'} = 1 if ( $masq =~ m/s/ );    # size
	$masq_array{'atime'}    = 1 if ( $masq =~ m/a/ );    # atime
	$masq_array{'mtime'}    = 1 if ( $masq =~ m/m/ );    # mtime
	$masq_array{'ctime'}    = 1 if ( $masq =~ m/c/ );    # ctime
	$masq_array{'blocs'}    = 1 if ( $masq =~ m/b/ );    # blocs
	   #	$masq_array{'sha'}      = 1 if ( $masq =~ m/[126]/ );    # sha checksum
	$masq_array{'sha1'}   = 1 if ( $masq =~ m/1/ );    # sha1 checksum
	$masq_array{'sha256'} = 1 if ( $masq =~ m/6/ );    # sha256 checksum
	$masq_array{'sha512'} = 1 if ( $masq =~ m/2/ );    # sha512 checksum

	return %masq_array;
}
#######################################################
# apply masq on fileinfo
sub apply_masq($$$) {
	my $rh_masq  = shift @_;                           # masq hash
	my $rh_info1 = shift @_;                           # fileinfo hash
	my $rh_info2 = shift @_;                           # fileinfo hash

	foreach my $key ( keys %{$rh_info1} ) {
		if ( !exists $rh_masq->{$key} ) {
			$rh_info1->{$key} = 0;
			$rh_info2->{$key} = 0;
		}
	}
	return;
}
#######################################################
# return true if the 2 fileinfo are different
# the attribute masq is used to masq unwanted fields
sub is_changed($$$) {
	my $self = shift @_;
	my $new  = shift @_;    # new fileinfo
	my $masq = shift @_;    # internal masq

	my %fileinfo_old = $self->to_hash();
	my %fileinfo_new = $new->to_hash();

	# transform masq in a masq array
	my %masq_array = masq2hash($masq);

	# apply masq
	apply_masq( \%masq_array, \%fileinfo_old, \%fileinfo_new );

	return !compare_hash( \%fileinfo_old, \%fileinfo_new );
}
#######################################################
# compare 2 file infos and print the difference
sub display_changes($$$$) {
	my $self   = shift @_;
	my $entry2 = shift @_;    # new value Afick::Object
	my $config = shift @_;    # Afick::Cfg
	my $report = shift @_;    # Afick::Report

	my $name = $self->name();
	my $masq = $report->get_masq($name);

	# transform masq in a masq array
	my %masq_array = masq2hash($masq);

	my $output = q{};
  FIELD: foreach my $key ( $self->fields() ) {

		if ( !defined $self->{$key} ) {
			Afick::Msg->warning("pb display_changes $name on old $key");
			next FIELD;
		}
		elsif ( !defined $entry2->{$key} ) {
			Afick::Msg->warning("pb display_changes $name on new $key");
			next FIELD;
		}
		elsif ( $self->{$key} ne $entry2->{$key} ) {

			my $field;
			if ( $config->get_directive('report_context') ) {

				# should make a difference between warning field and info fields
				# using masks
				# cf Log.pm
				if ( $masq_array{$key} ) {

					# real warning attribute : add w_ prefix
					$field = 'w_' . $key;
				}
				else {
					# info attribute : add i_ prefix
					$field = 'i_' . $key;
				}
			}
			else {
				# only report attributes from rules (warning)
				if ( $masq_array{$key} ) {
					$field = $key;
				}
				else {
					# print "skip $key (not in masq)\n";
					next FIELD;
				}
			}

			# !!!! for windows acl, it is long and difficult to read
			# so a special report is necessary
			if ( ( $key eq 'gid' ) and is_acl() ) {
				Afick::WinAcl::compare_acl( $self->gid(), $entry2->gid(),
					$report );
			}
			else {

				# general case
				$output .= $report->report_detailed(
					$field,
					$self->display($key),
					$entry2->display($key)
				);
			}
		}    # if fileinfo ne
	}    # for
	return $output;
}
#######################################################
sub check($) {
	my $self = shift @_;

	# recode old sha field into sha1 (guess)
	if ( exists $self->{'sha'} ) {
		if ( !exists $self->{'sha1'} ) {
			$self->{'sha1'} = $self->{'sha'};
		}
		delete $self->{'sha'};
		Afick::Msg->info(
			'convert sha field into sha1 for ' . $self->{'name'} );
	}
	return;
}
#######################################################

1;
