#!/usr/bin/env perl -T
###############################################################################
#
#    Copyright (C) 2002-2023 by Eric Gerbier
#    Bug reports to: eric.gerbier@tutanota.com
#    $Id$
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
###############################################################################
# this class is designed to implement a way to access package database data
# this is a wrapper to specific databases : rpm, deb ...
###############################################################################
# I use special naming for references :
# $r_name : for a ref to scalar
# $rh_name : for a ref to hashes
# $ra_name : for a ref to array

package Afick::Package;

use strict;
use warnings;

use English qw(-no_match_vars);

use base qw(Exporter);
## no critic (ProhibitAutomaticExportation)
our ( $VERSION, @EXPORT, );
$VERSION = '1.0';
@EXPORT  = qw( );

# optional modules
# ###################

# other afick modules
# #####################
use Afick::Constant;
use Afick::Msg;
use Afick::Rpm;
use Afick::Deb;

###############################################################
# constructor
sub new($) {
	my $classe = shift @_;

	my $self = {};
	bless $self, $classe;
	$self->_init();
	return $self;
}

###############################################################
# try to find a package manager
sub _init($) {
	my $self = shift @_;

	# internal cache
	# because it is called 2 times in report summary and report detailed
	$self->{'cache'} = undef;

	# rpm or deb
	$self->{'package'} = undef;

	my @list = ( 'Afick::Rpm', 'Afick::Deb' );
	foreach my $elem (@list) {
		my $pac = $elem->new();
		if ( $pac->check() ) {

			$self->{'package'} = $pac;
			last;
		}
	}
	return;
}
################################################################
sub is_initialised($) {
	my $self = shift @_;

	return defined $self->{'package'};
}
################################################################
# fhs moves some directories in /usr
# /bin => /usr/bin
# /sbin => /usr/sbin
# /lib => /usr/lib
# /lib64 => /usr/lib64
# but packages in debian/ubuntu have 2 problems :
# - some uses /lib, some uses /usr/lib
# - dpkg -S does not follow the links
# example :
# /bin/dpkg -S /usr/lib/modules/5.19.0-31-generic/kernel/drivers/net/can/softing/softing.ko
#   dpkg-query: no path found matching pattern /usr/lib/modules/5.19.0-31-generic/kernel/drivers/net/can/softing/softing.ko
# /bin/dpkg -S /lib/modules/5.19.0-31-generic/kernel/drivers/net/can/softing/softing.ko
#   linux-modules-extra-5.19.0-31-generic: /lib/modules/5.19.0-31-generic/kernel/drivers/net/can/softing/softing.ko
#
# dpkg -S /usr/lib/libcpufreq.so.0.0.0
#   libcpufreq0: /usr/lib/libcpufreq.so.0.0.0
# dpkg -S /lib/libcpufreq.so.0.0.0
#   dpkg-query: no path found matching pattern /lib/libcpufreq.so.0.0.0
#
# so we have to check for each file the 2 paths
sub fhs($$) {
	my $self     = shift @_;
	my $filename = shift @_;

	my @list_dir = ( 'bin', 'sbin', 'lib', 'lib64' );

	my @list_fhs;
	push @list_fhs, $filename;
	foreach my $dir (@list_dir) {
		if ( $filename =~ m{^/usr/$dir/(.*)} ) {
			push @list_fhs, "/$dir/$1";
		}
		elsif ( $filename =~ m{^/$dir/(.*)} ) {
			push @list_fhs, "/usr/$dir/$1";
		}
	}
	return @list_fhs;
}
################################################################
# from a filename, search the package name
sub file2pac($$) {
	my $self     = shift @_;
	my $filename = shift @_;

	if ( $self->is_initialised() ) {
		my @list_fhs = $self->fhs($filename);
		foreach my $fic (@list_fhs) {
			my $pac = $self->{'package'}->file2pac($fic);
			return $pac if $pac;
		}
	}
	return;
}
################################################################
# check if the give filename is ok (unchanged) in package manager
# database
# called from Afick::Report print_new print_changed
# the cache is used to avoid many calls to package program
# the fist algo was very slow on kernel changes !
# we assume that if a package is updated, many files are changed
# with this algo, we have only 3 packages calls, instead of one/two by file
# 1 file2pac
# 2 list_chaged_files
# 3 list_pac_files
# for next detected file, all is in cache
sub is_dbconform($$) {
	my $self     = shift @_;
	my $filename = shift @_;

	return unless $filename;
	if ( $self->is_initialised() ) {
		if ( exists $self->{'cache'}{$filename} ) {
			Afick::Msg->debug("(is_dbconform) use cache for $filename");
			return $self->{'cache'}{$filename};
		}
		else {
			my $pac = $self->{'package'}->file2pac($filename);
			if ($pac) {
				Afick::Msg->debug("(is_dbconform) file2pac $filename => $pac");
				my %h_change = $self->{'package'}->list_changed_files($pac);
				my @list_pac = $self->{'package'}->list_pac_files($pac);

				# put all package files in cache
				foreach my $fic (@list_pac) {
					my @liste_fhs = $self->fhs($fic);
					foreach my $fhs (@liste_fhs) {

						#Afick::Msg->debug("(is_dbconform) put $fhs in cache");
						$self->{'cache'}{$fhs} = ( !exists $h_change{$fhs} );
					}
				}
			}
			else {
				$self->{'cache'}{$filename} = 0;
			}
			return $self->{'cache'}{$filename};
		}
	}
	else {
		# can not check
		# return false
		return;
	}
}
################################################################

1;
__END__

=head1 NAME

Afick::Package

=head1 DESCRIPTION

=head1 DIAGNOSTICS

=head1 USAGE

use Afick::Package;

my $pac = Afick::Package->new();

$pac->is_dbconform('foo');

=head1 OPTIONS

=head1 REQUIRED ARGUMENTS

=head1 EXIT STATUS

=head1 CONFIGURATION

=head1 DEPENDENCIES

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 LICENSE AND COPYRIGHT

Copyright (c) 2002 Eric Gerbier
All rights reserved.

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option)
any later version.

=head1 AUTHOR

Eric Gerbier

you can report any bug or suggest to eric.gerbier@tutanota.com

