#!/usr/bin/env perl -T
###############################################################################
#
#    Copyright (C) 2002-2023 by Eric Gerbier
#    Bug reports to: eric.gerbier@tutanota.com
#    $Id$
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
###############################################################################
# this classe is designed to implement package database access
###############################################################################
# I use special naming for references :
# $r_name : for a ref to scalar
# $rh_name : for a ref to hashes
# $ra_name : for a ref to array

package Afick::Rpm;

use strict;
use warnings;

use English qw(-no_match_vars);

use base qw(Exporter);
## no critic (ProhibitAutomaticExportation)
our ( $VERSION, @EXPORT, );
$VERSION = '1.1';
@EXPORT  = qw( );

# optional modules
# ###################

# other afick modules
# #####################
use Afick::Constant;
use Afick::Msg;
use Afick::Gen;    # search_path

my $Dbcmd   = 'rpm';                  # rpm command and rpm package name
my $F_dbcmd = search_path($Dbcmd);    # full path to rpm command

## no critic (ProhibitBacktickOperators)
###############################################################
# constructor
sub new($) {
	my $classe = shift @_;

	my $self = {};
	bless $self, $classe;
	$self->_init();
	return $self;
}

###############################################################
# reset internal variables
sub _init($) {
	my $self = shift @_;

	# internal cache
	#$self->{'cache'} = undef;

	return;
}
################################################################
# is used to check if this database is installed and in use
sub check($) {
	my $self = shift @_;

	# check if rpm is installed
	if ($F_dbcmd) {

		# check if rpm database is filled
		my $out = `$F_dbcmd -q $Dbcmd `;

		# to avoid translation problem
		# package rpm is not installed
		return if ( $out =~ m/ / );
		return $out;
	}
	else {
		# command not found
		return;
	}
}
################################################################
# convert a filename to a package name
sub file2pac($$) {
	my $self     = shift @_;
	my $filename = shift @_;

	return unless ($F_dbcmd);
	return unless ($filename);
	return unless ( -e $filename );

	Afick::Msg->debug("(file2pac) $filename");

	# in theory a file should belong to only a package or none
	# but not in real life
	# for example : /usr/share/licenses/kernel-core
	# rpm -qf /usr/share/licenses/kernel-core
	# kernel-core-6.1.8-200.fc37.x86_64
	# kernel-core-6.1.9-200.fc37.x86_64
	# kernel-core-6.1.10-200.fc37.x86_64
	my $cmd  = "$F_dbcmd -qf $filename";
	my @pacs = `$cmd`;
	my $pac  = $pacs[0];                   # take first
	return unless $pac;
	chomp $pac;

	# file /mnt/data/projets/afick/src/Rpm.pm is not owned by any package
	# package name does not contains spaces
	return if ( $pac =~ m/ / );

	return $pac;
}
################################################################
# return the list of files in the given package
sub list_pac_files($$) {
	my $self = shift @_;
	my $pac  = shift @_;

	return unless ($F_dbcmd);
	return unless $pac;
	Afick::Msg->debug("(list_pac_files) $pac");

	my @list = `$F_dbcmd -ql $pac`;
	chomp @list;
	return @list;
}
################################################################
# return a hash of modified files against package database
sub list_changed_files($$) {
	my $self = shift @_;
	my $pac  = shift @_;

	return unless ($F_dbcmd);
	Afick::Msg->debug("(list_changed_files) $pac");

	my %h;
	my $cmd = "$F_dbcmd -V $pac";

	# S.?....T.  c /etc/afick.conf
	# S.5....T.    /usr/bin/afick-common.pl
	## no critic (ProhibitTwoArgOpen)
	if ( open my $fh, "$cmd |" ) {
		while (<$fh>) {
			my @tab      = split q{ };
			my $changes  = $tab[0];
			my $filename = $tab[-1];
			chomp $filename;
			$h{$filename} = $changes;
		}
		## no critic (RequireCheckedClose,RequireCheckedSyscalls)
		close $fh;
	}
	else {
		Afick::Msg->warning("(verify_pac) can not open $cmd : $ERRNO");
	}
	return %h;
}
################################################################

1;
__END__

=head1 NAME

Afick::Rpm

=head1 DESCRIPTION

=head1 DIAGNOSTICS

=head1 USAGE

use Afick::Rpm;

my $pm = Afick::Rpm->new();

$res = $pm->file2pac('/etc/fstab');

my %h = $pm->verify_pac(bash);

=head1 OPTIONS

=head1 REQUIRED ARGUMENTS

=head1 EXIT STATUS

=head1 CONFIGURATION

=head1 DEPENDENCIES

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 LICENSE AND COPYRIGHT

Copyright (c) 2002 Eric Gerbier
All rights reserved.

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option)
any later version.

=head1 AUTHOR

Eric Gerbier

you can report any bug or suggest to eric.gerbier@tutanota.com

