#!/usr/bin/env perl -T
###############################################################################
#
#    Copyright (C) 2002-2004 by Eric Gerbier
#    Bug reports to: eric.gerbier@tutanota.com
#    $Id$
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
###############################################################################
# a general package for all acl routines
###############################################################################
# case problem :
# - masq is a number : no problem
# - rights are a list constants : no problem
# - sid : case seems to vary
# we can not force case just in compare, because is_change works on stored sid/masq
###############################################################################
# I use special naming for references :
# $r_name : for a ref to scalar
# $rh_name : for a ref to hashes
# $ra_name : for a ref to array

package Afick::WinAcl;

use strict;
use warnings;

use English qw(-no_match_vars);
use base qw(Exporter);
use File::Basename;

use Win32::FileSecurity;
use Afick::Msg;
use Afick::Tst;    # is_acl
use Afick::Report;

## no critic (ProhibitAutomaticExportation)
our ( $VERSION, @EXPORT, );

$VERSION = '0.3';
@EXPORT  = qw( &split_acl &winacl &hash_acl &compare_acl );

my $Sepacl = q{,};
#######################################################
# acl
#######################################################
# get windows acl for a given file
# return sid/masq instead sid/right to reduce record size
sub winacl($$) {
	my $filename = shift @_;    # file name
	my $dbm      = shift @_;

	my $acl = 0;
	my %perm;
	if ( is_acl() ) {
		my @acl;

		# catch the error to avoid stop the program
		eval { Win32::FileSecurity::Get( $filename, \%perm ) };
		if ($EVAL_ERROR) {
			Afick::Msg->debug(
				"(winacl) can not get windows acl for $filename : $EVAL_ERROR",
				3
			);
		}
		else {
			while ( my ( $sid, $mask ) = each %perm ) {
				$sid =~ s/ /_/g;
				$sid = lc $sid;    # force sid in lower case
				push @acl, "$sid=$mask";
			}
			$acl = join $Sepacl, sort @acl;

	   # caution : acl may be big, too big for sdbm database :
	   # There are a number of limits on the size of the data that you can store
	   # in the SDBM file. The most important is that the length of a key, plus
	   # the length of its associated value, may not exceed 1008 bytes.
			if ( $dbm eq 'SDBM_File' ) {
				## no critic (ProhibitParensWithBuiltins)
				my $field_len = 210;    # checksum and others fields
				my $record_len = $field_len + length($filename) + length($acl);
				if ( $record_len > 1008 ) {
					Afick::Msg->warning(
"(winacl) remove acl for $filename, was too long : $record_len"
					);

					# what information to record ?
					# one way is to replace acl by a warning
					$acl = 'removed (too long)';

					# one way may be to truncate acl,
					# but must truncate on Sepacl to avoid decoding problem
					# my $max_acl_len   = 1008 - $field_len - length($filename);
					# my $index         = rindex( $acl, $Sepacl, $max_acl_len );
					# my $truncated_acl = substr( $acl, 0, $index );
					# $acl = $truncated_acl;
				}
			}
		}
	}
	return $acl;
}
#######################################################
# transform a windows acl from array (split_acl)
# to a hash, to allow a easy compare
sub hash_acl(@) {
	my @tab = @_;

	my %acl;
	foreach my $elem (@tab) {
		my ( $sid, $rights ) = split /=/, $elem;
		$acl{$sid} = $rights;
	}
	return %acl;
}
#######################################################
# transform a window's acl
# from binary form in hashtable
# to human way for display
sub split_acl($) {
	my $acl = shift @_;    # windows acl

	my @tab_acl = split /$Sepacl/, $acl;
	my @text;
	foreach my $acl_item (@tab_acl) {
		my ( $sid, $mask ) = split /=/, $acl_item;

		next if ( !defined $mask );

		# force all data in lower case
		my @rights;
		Win32::FileSecurity::EnumerateRights( $mask, \@rights );
		push @text, "$sid=" . join $Sepacl, @rights;
	}
	return @text;
}
#######################################################
# compare 2 acl and display new, deleted, changed
sub compare_acl($$$) {
	my $acl_old = shift @_;    # old value from database
	my $acl_new = shift @_;    # new value from database
	my $report  = shift @_;    # Afick::Report object

	my %acl_old = hash_acl( split_acl($acl_old) );
	my %acl_new = hash_acl( split_acl($acl_new) );

	my $nb_change = 0;
	my $nb_delete = 0;
	my $nb_new    = 0;
	foreach my $sid ( keys %acl_old ) {
		if ( exists $acl_new{$sid} ) {
			if ( $acl_old{$sid} ne $acl_new{$sid} ) {

				# changed acl
				$report->report_detailed( "change acl $sid",
					$acl_old{$sid}, $acl_new{$sid} );
				$nb_change++;
			}
			delete $acl_new{$sid};
		}
		else {

			# deleted acl
			$report->report_detailed( "deleted acl $sid", $acl_old{$sid} );
			$nb_delete++;
		}
	}

	# we have deleted all common keys on acl_new
	# so we just have new acl now
	foreach my $sid ( keys %acl_new ) {

		# new acl
		$report->report_detailed( "new acl $sid", $acl_new{$sid} );
		$nb_new++;
	}
	return ( $nb_new, $nb_delete, $nb_change );
}
#######################################################
1;
